//=============================================================================
// Serial Communication Class Library
// Copyright(c)1999 Yamamoto Takanori.
//=============================================================================
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "ezring.h"
#include "serial.h"

void serial_nop(void)
{
  // NOP
}

int serial::init(int bps)
{
  int i;
  struct termios term;

  /* Set the tty to raw and to the correct speed */
  if(tcgetattr(fd, &term)==-1){return -1; /* GET_ERR */}
  term.c_oflag = 0;
  term.c_iflag = IGNBRK | IGNPAR;
  term.c_cflag = CREAD | CLOCAL | CS8;
  cfsetispeed(&term, bps); cfsetospeed(&term, bps);
  term.c_lflag = NOFLSH;

  /* cfmakeraw() */
  term.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
  term.c_oflag &= ~OPOST;
  term.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
  term.c_cflag &= ~(CSIZE|PARENB);
  term.c_cflag |= CS8;

  for(i=0;i<16;i++){term.c_cc[i]=0;}
  term.c_cc[VMIN] = 1;
  term.c_cc[VTIME] = 0;
  if(tcsetattr(fd, TCSANOW, &term)==-1){return -2; /* SET_ERR */}

  if ((i = fcntl(fd, F_GETFL, 0))!=-1) {
    i &= ~O_NONBLOCK;
    fcntl(fd, F_SETFL, i);
  }

  return 0;
}

//CONSTRUCTOR
serial::serial(int bufsize)
{
  ring.setbufsize(bufsize);
  ms=0; //default timeout is disabled.
  flag_timeout=0;
  fd=-1;
  setcallbackfunc_timeout(serial_nop);
}

//DESTRUCTOR
serial::~serial()
{
  if(fd>=0){close();}
}

int serial::open(const char *siodev, int bps)
{
  int err;
  if((fd = ::open(siodev, O_RDWR | O_NONBLOCK)) < 0){
    return -1; // OPEN_ERR
  }
  if((err = init(bps)) != 0){
    if(err == -1){
      return -2; // INIGET_ERR
    }else if(err == -2){
      return -3; // INISET_ERR
    }
    return -4; // UNK_ERR 
  }
  return 0;
}

int serial::close(void)
{
  ::close(fd);
  return 0;
}

int serial::read(char *str, int len)
{
  int fdx;
  fd_set rset;
  int n;
  char *buff;
  int sizeofbuff = ring.getbufsize()/4;
  flag_timeout=0;

  buff=(char *)malloc(sizeofbuff);

  FD_ZERO(&rset);
  FD_SET(fd , &rset); fdx = fd+1;

  while(1){
    int ret;
    if(ring.length()>=len){break;}

    if(ms==0){
      ret = select(fdx, &rset, NULL, NULL, NULL);
    }else{
      struct timeval timeout;
      timeout.tv_sec =ms / 1000;          // seconds
      timeout.tv_usec=ms % 1000;          // microseconds
      ret = select(fdx, &rset, NULL, NULL, &timeout);
    }
    if(ret==0){//TIMEOUT
      callbackfunc_timeout();
      flag_timeout=1;
      break;
    }
    if(FD_ISSET(fd, &rset)){
        n = ::read(fd, buff, sizeofbuff);
        ring.append(buff,n);
    }
  }
  free(buff);
  return(ring.nextract(str, len));
}

int serial::read_until(char *str, int len, char c)
{
// this function is perhaps very robust, but very slowly...
  int n=0;
  flag_timeout=0;
  while(1){
    if(n>=len){return n;}
    this->read((str+n), 1);
    if(flag_timeout){  //TIMEOUT OCCURRED
      return n;
    }
    if(str[n] == c){
      return (n+1);
    }else{
      n++;
    }
  }
}

int serial::write(const char *str, int len)
{
  int m, n=0;
  while(n<len){
    if((m = ::write(fd, str, len))<0){
      break; /* write error */
    }
    n+=m;
  }
  return 0;
}

int serial::setbufsize(int size)
{
  ring.setbufsize(size);
  return 0;
}


int serial::getbufsize(void)
{
  return ring.getbufsize();
}


int serial::settimeout(int val)
{
  ms=val;
  return 0;
}

int serial::gettimeout(void)
{
  return ms;
}

int serial::setcallbackfunc_timeout(void (*f)(void))
{
  callbackfunc_timeout = f;
  return 0;
}
